searchData={"items":[{"type":"module","title":"Telemetry.Metrics","doc":"Common interface for defining metrics based on\n[`:telemetry`](https://github.com/beam-telemetry/telemetry) events.\n\nMetrics are aggregations of Telemetry events with specific names, providing\na view of the system's behaviour over time.\n\nTo give a more concrete example, imagine that somewhere in your code there is\na function which sends an HTTP request, measures the time it took to get a\nresponse, and emits an event with the information:\n\n    :telemetry.execute([:http, :request, :stop], %{duration: duration})\n\nYou could define a counter metric, which counts how many HTTP requests were\ncompleted:\n\n    Telemetry.Metrics.counter(\"http.request.stop.duration\")\n\nor you could use a summary metric to see statistics about the request duration:\n\n    Telemetry.Metrics.summary(\"http.request.stop.duration\")\n\nThis documentation is going to cover all the available metrics and how to use\nthem, as well as options, and how to integrate those metrics with reporters.","ref":"Telemetry.Metrics.html"},{"type":"module","title":"Metrics - Telemetry.Metrics","doc":"There are five metric types provided by `Telemetry.Metrics`:\n\n  * `counter/2` which counts the total number of emitted events\n  * `sum/2` which keeps track of the sum of selected measurement\n  * `last_value/2` holding the value of the selected measurement from\n    the most recent event\n  * `summary/2` calculating statistics of the selected measurement,\n    like maximum, mean, percentiles etc.\n  * `distribution/2` which builds a histogram of selected measurement\n\nThe first argument to all metric functions is the metric name. Metric\nname can be provided as a string (e.g. `\"http.request.stop.duration\"`) or a\nlist of atoms (`[:http, :request, :stop, :duration]`). The metric name is\nautomatically used to infer the telemetry event and measurement. For example,\nIn the `\"http.request.stop.duration\"` example, the source event name is\n`[:http, :request, :stop]` and metric values are drawn from `:duration`\nmeasurement. Like this:\n\n    [:http , :request, :stop]      :duration\n    <----- event name ------> <-- measurement -->\n\nYou can also explicitly specify the event name and measurement\nif you prefer.\n\nThe second argument is a list of options. Below is the description of the\noptions common to all metric types:\n\n  * `:event_name` - the source event name. Can be represented either as a\n    string (e.g. `\"http.request\"`) or a list of atoms (`[:http, :request]`).\n    By default the event name is all but the last segment of the metric name.\n  * `:measurement` - the event measurement used as a source of a metric values.\n    By default it is the last segment of the metric name. It can be:\n    * a key in the event's measurements map\n    * an unary function accepting the whole measurements map and returning the\n      actual value to be used.\n    * a binary function accepting the measurements and metadata maps and\n      returning the actual value to be used.\n  * `:tags` - a subset of metadata keys by which aggregations will be broken down.\n    Defaults to an empty list.\n  * `:tag_values` - a function that receives the metadata and returns a map with\n    the tags as keys and their respective values. Defaults to returning the\n    metadata itself.\n  * `:keep` - a predicate function that evaluates the metadata to conditionally\n    record a given event. `:keep` and `:drop` cannot be combined. Defaults to `nil`.\n  * `:drop` - a predicate function that evaluates the metadata to conditionally\n    skip recording a given event. `:keep` and `:drop` cannot be combined. Defaults to `nil`.\n  * `:description` - human-readable description of the metric. Might be used by\n    reporters for documentation purposes. Defaults to `nil`.\n  * `:unit` - an atom describing the unit of selected measurement, typically in\n    singular, such as `:millisecond`, `:byte`, `:kilobyte`, etc. It may also be\n    a tuple indicating that a measurement should be converted from one unit to\n    another before a metric is updated. Currently, only time and byte unit\n    conversions are supported. We discuss those in detail in the \"Converting Units\"\n    section.\n  * `:reporter_options` - a keyword list of reporter-specific options for the metric.","ref":"Telemetry.Metrics.html#module-metrics"},{"type":"module","title":"Breaking down metric values by tags - Telemetry.Metrics","doc":"Sometimes it's not enough to have a global overview of all HTTP requests received\nor all DB queries made. It's often more helpful to break down this data, for example,\nwe might want to have separate metric values for each unique database table and\noperation name (`select`, `insert` etc.) to see how these particular queries behave.\n\nThis is where tagging comes into play. All metric definitions accept a `:tags` option:\n\n    counter(\"db.query.duration\", tags: [:table, :operation])\n\nThe above definition means that we want to keep track of the number of queries, but\nwe want a separate counter for each unique pair of table and operation. Tag values are\nfetched from event metadata - this means that in this example, `[:db, :query]` events\nneed to include `:table` and `:operation` keys in their payload:\n\n    :telemetry.execute([:db, :query], %{duration: 198}, %{table: \"users\", operation: \"insert\"})\n    :telemetry.execute([:db, :query], %{duration: 112}, %{table: \"users\", operation: \"select\"})\n    :telemetry.execute([:db, :query], %{duration: 201}, %{table: \"sessions\", operation: \"insert\"})\n    :telemetry.execute([:db, :query], %{duration: 212}, %{table: \"sessions\", operation: \"insert\"})\n\nThe result of aggregating the events above looks like this:\n\n| table    | operation | count |\n| -------- | --------- | ----- |\n| users    | insert    | 1     |\n| users    | select    | 1     |\n| sessions | insert    | 2     |\n\nThe approach where we create a separate metric for some unique set of properties\nis called a multi-dimensional data model.\n\n#","ref":"Telemetry.Metrics.html#module-breaking-down-metric-values-by-tags"},{"type":"module","title":"Transforming event metadata for tagging - Telemetry.Metrics","doc":"Finally, sometimes there is a need to modify event metadata before it's used for\ntagging. Each metric definition accepts a function in `:tag_values` option which\ntransforms the metadata into desired shape. Note that this function is called for\neach event, so it's important to keep it fast if the rate of events is high.","ref":"Telemetry.Metrics.html#module-transforming-event-metadata-for-tagging"},{"type":"module","title":"Converting Units - Telemetry.Metrics","doc":"It might happen that the unit of measurement we're tracking is not the desirable unit\nfor the metric values, e.g. events are emitted by a 3rd-party library we do not control,\nor a reporter we're using requires specific unit of measurement.\n\nFor these scenarios, each metric definition accepts a `:unit` option in a form of a tuple:\n\n    summary(\"http.request.stop.duration\", unit: {from_unit, to_unit})\n\nThis means that the measurement will be converted from `from_unit` to `to_unit` before\nbeing used for updating the metric. Currently, only time and byte conversions are\nsupported.\n\n#","ref":"Telemetry.Metrics.html#module-converting-units"},{"type":"module","title":"Time Conversions - Telemetry.Metrics","doc":"Most time measurements in the Erlang VM are done in the `:native` unit, which we need\nto convert to the desired precision. The supported time units are: `:second`, `:millisecond`,\n`:microsecond`, `:nanosecond` and `:native`.\n\nFor example, to convert HTTP request duration from `:native` time unit to milliseconds\nyou'd write:\n\n    summary(\"http.request.stop.duration\", unit: {:native, :millisecond})\n\n#","ref":"Telemetry.Metrics.html#module-time-conversions"},{"type":"module","title":"Byte Conversions - Telemetry.Metrics","doc":"Some metrics, like VM memory's usage are reported in bytes. You might want to convert this\nto megabytes, for example. The supported byte units are: `:byte`, `:kilobyte` and `:megabyte`.\n\nIn order to convert a metric value from bytes to megabytes, you can write the following:\n\n    last_value(\"vm.memory.total\", unit: {:byte, :megabyte})","ref":"Telemetry.Metrics.html#module-byte-conversions"},{"type":"module","title":"VM metrics - Telemetry.Metrics","doc":"`Telemetry.Metrics` doesn't have a special treatment for the VM metrics - they need\nto be based on the events like all other metrics.\n\n`:telemetry_poller` package (http://hexdocs.pm/telemetry_poller) exposes a bunch of\nVM-related metrics and also provides custom periodic measurements. You can add\ntelemetry poller as a dependency:\n\n    {:telemetry_poller, \"~> 0.4\"}\n\nBy simply adding `:telemetry_poller` as a dependency, two events will become available:\n\n  * `[:vm, :memory]` - contains the total memory, as well as the memory used for\n    binaries, processes, etc. See `erlang:memory/0` for all keys;\n  * `[:vm, :total_run_queue_lengths]` - returns the run queue lengths for CPU and\n    IO schedulers. It contains the `total`, `cpu` and `io` measurements;\n\nYou can consume those events with `Telemetry.Metrics` with the following sample metrics:\n\n    last_value(\"vm.memory.total\", unit: :byte)\n    last_value(\"vm.total_run_queue_lengths.total\")\n    last_value(\"vm.total_run_queue_lengths.cpu\")\n    last_value(\"vm.total_run_queue_lengths.io\")\n\nIf you want to change the frequency of those measurements, you can set the\nfollowing configuration in your config file:\n\n    config :telemetry_poller, :default, period: 5_000 # the default\n\nOr disable it completely with:\n\n    config :telemetry_poller, :default, false\n\nThe `:telemetry_poller` package also allows you to run your own poller, which is\nuseful to retrieve process information or perform custom measurements periodically.\nFor example, to keep track of the number of users, inside a supervision tree, you could do:\n\n    measurements = [\n      {:process_info,\n       event: [:my_app, :worker],\n       name: MyApp.Worker,\n       keys: [:message_queue_len, :memory]},\n\n      {MyApp, :measure_users, []}\n    ]\n\n    Supervisor.start_link([\n      # Run the given measurements every 10 seconds\n      {:telemetry_poller, measurements: measurements(), period: 10_000}\n    ], strategy: :one_for_one)\n\nWhere `MyApp.measure_users/0` could be written like this:\n\n    defmodule MyApp do\n      def measure_users do\n        :telemetry.execute([:my_app, :users], %{total: MyApp.users_count()}, %{})\n      end\n    end\n\nNow with measurements in place, you can define the metrics for the\nevents above:\n\n    last_value(\"my_app.worker.memory\", unit: :byte)\n    last_value(\"my_app.worker.message_queue_len\")\n    last_value(\"my_app.users.total\")","ref":"Telemetry.Metrics.html#module-vm-metrics"},{"type":"module","title":"Optionally Recording Metric Events - Telemetry.Metrics","doc":"There may be occasions where you want to record metrics differently or not at all based\nupon metadata. Rather than depending on changing event names with prefixes, you can instead\nprovide a predicate function which returns true when you want the metric to be\nprocessed for this event and false when you do not.\n\nLet's examine some a few examples where optional recording can be helpful.\n\n#","ref":"Telemetry.Metrics.html#module-optionally-recording-metric-events"},{"type":"module","title":"Filtering on Metadata - Telemetry.Metrics","doc":"Let's assume you are using an HTTP client library in your application which has the following\nevent_name: `[:http_client, :request, :stop]`. You use this library in multiple places and\nyou'd like to monitor the request duration.\n\nYou can use the event provided by the library but you have very different acceptable\nperformance requirements for a critical request, so it would be better to provide a different\nmetric name for monitoring. The client library includes a user configured `:name` option which\nyou can set and is passed in the event metadata.\n\nLet's create a default distribution metric and one for the high performance call.\n\n    distribution(\n      \"http.client.request.duration\",\n      event_name: [:http_client, :request, :stop],\n      drop: &(match?(%{name: :fast_client}, &1))\n    )\n\n    distribution(\n      \"http.fast.client.request.duration\",\n      event_name: [:http_client, :request, :stop],\n      keep: &(match?(%{name: :fast_client}, &1))\n    )\n\nWith this configuration, you can now monitor these requests separately. In the first example,\nany events where the client's name is `:fast_client` will be dropped for that metric.\nConversely, any matching events in the second example metric will be kept.\n\nNote: only `:keep` OR `:drop` may be set, never both.\n\n#","ref":"Telemetry.Metrics.html#module-filtering-on-metadata"},{"type":"module","title":"Other Uses - Telemetry.Metrics","doc":"Another potential use case for `:keep | :drop` could be per metric sampling rates. As long\nas your function returns a boolean, it can determine if the event should be processed.\n\n#","ref":"Telemetry.Metrics.html#module-other-uses"},{"type":"module","title":"Reporter Support - Telemetry.Metrics","doc":"Event optional recording must be supported by the reporter you are using. Check your reporter's\ndocumentation before relying on this functionality.\n\nThe `keep` function should be evaluated by reporters _prior_ to `tag_values` using the\nraw `:telemetry.medatadata()` values from the event.","ref":"Telemetry.Metrics.html#module-reporter-support"},{"type":"module","title":"Reporters - Telemetry.Metrics","doc":"So far, we have talked about metrics and how to describe them, but we haven't discussed\nhow those metrics are consumed and published to a system that provides data visualization,\naggregation, and more. The job of subscribing to events and processing the actual metrics\nis a responsibility of reporters.\n\nGenerally speaking, a reporter is a process that you would start in your supervision\ntree with a list of metrics as input. For example, `Telemetry.Metrics` ships with a\n`Telemetry.Metrics.ConsoleReporter` module, which prints data to the terminal as an\nexample. You would start it as follows:\n\n    metrics = [\n      last_value(\"my_app.worker.memory\", unit: :byte),\n      last_value(\"my_app.worker.message_queue_len\"),\n      last_value(\"my_app.users.total\")\n    ]\n\n    Supervisor.start_link([\n      {Telemetry.Metrics.ConsoleReporter, metrics: metrics}\n    ], strategy: :one_for_one)\n\nReporters take metric definitions as an input, subscribe to relevant events and\naggregate data when the events are emitted. Reporters may push metrics to StatsD,\nsome time-series database, or exposing a HTTP endpoint for Prometheus to scrape.\nIn a nutshell, `Telemetry.Metrics` defines only how metrics of particular type\nshould behave and reporters provide the actual implementation for these aggregations.\n\nOfficial reporters, maintained by the Observability Working Group of the Erlang\nEcosystem Foundation, can be found on the [BEAM Telemetry organization on GitHub](https://github.com/beam-telemetry/).\nYou may also [find community reporters on hex.pm](https://hex.pm/packages?search=telemetry_metrics).\nYou can also read the [Writing Reporters](writing_reporters.html) page for general\ninformation on how to write a reporter.","ref":"Telemetry.Metrics.html#module-reporters"},{"type":"module","title":"Wiring it all up - Telemetry.Metrics","doc":"Over the previous sections we discussed how to setup metrics and pass them to reporters\nand how to configure a poller for measurements. We can wire it all up into a single\nmodule as shown below. The example below would be used in the context of a Phoenix\napplication, where we have web metrics, database metrics (through Ecto) as well as\nfrom the database, Phoenix metrics as well as VM metrics.\n\nThe first step is to add both `:telemetry_metrics` and `:telemetry_poller` as\ndependencies:\n\n    [\n      {:telemetry_poller, \"~> 0.4\"},\n      {:telemetry_metrics, \"~> 0.4\"}\n    ]\n\nThen you could define a module that wires everything up:\n\n    defmodule MyAppWeb.Telemetry do\n      use Supervisor\n      import Telemetry.Metrics\n\n      def start_link(arg) do\n        Supervisor.start_link(__MODULE__, arg, name: __MODULE__)\n      end\n\n      def init(_arg) do\n        children = [\n          {:telemetry_poller,\n           measurements: periodic_measurements(),\n           period: 10_000},\n          # Or TelemetryMetricsPrometheus or TelemetryMetricsFooBar\n          {TelemetryMetricsStatsd, metrics: metrics()}\n        ]\n\n        Supervisor.init(children, strategy: :one_for_one)\n      end\n\n      defp metrics do\n        [\n          # VM Metrics\n          last_value(\"vm.memory.total\", unit: :byte),\n          last_value(\"vm.total_run_queue_lengths.total\"),\n          last_value(\"vm.total_run_queue_lengths.cpu\"),\n          last_value(\"vm.total_run_queue_lengths.io\"),\n\n          last_value(\"my_app.worker.memory\", unit: :byte),\n          last_value(\"my_app.worker.message_queue_len\"),\n\n          # Database Time Metrics\n          summary(\"my_app.repo.query.total_time\", unit: {:native, :millisecond}),\n          summary(\"my_app.repo.query.decode_time\", unit: {:native, :millisecond}),\n          summary(\"my_app.repo.query.query_time\", unit: {:native, :millisecond}),\n          summary(\"my_app.repo.query.idle_time\", unit: {:native, :millisecond}),\n          summary(\"my_app.repo.query.queue_time\", unit: {:native, :millisecond}),\n\n          # Phoenix Time Metrics\n          summary(\"phoenix.endpoint.stop.duration\",\n                  unit: {:native, :millisecond}),\n          summary(\n            \"phoenix.router_dispatch.stop.duration\",\n            unit: {:native, :millisecond},\n            tags: [:plug]\n          )\n        ]\n      end\n\n      defp periodic_measurements do\n        [\n          {:process_info,\n           event: [:my_app, :worker],\n           name: Rumbl.Worker,\n           keys: [:message_queue_len, :memory]}\n        ]\n      end\n    end","ref":"Telemetry.Metrics.html#module-wiring-it-all-up"},{"type":"function","title":"Telemetry.Metrics.counter/2","doc":"Returns a definition of counter metric.\n\nCounter metric keeps track of the total number of specific events emitted.\n\nThe value of the counter is always incremented by one, regardless of the\nvalue of the measurement. However, note the measurement must still be\navailable in the event, otherwise the event is not accounted for.\n\nSee the [\"Metrics\"](#module-metrics) section in the top-level documentation of this module for more\ninformation.","ref":"Telemetry.Metrics.html#counter/2"},{"type":"function","title":"Example - Telemetry.Metrics.counter/2","doc":"counter(\n      \"http.request.count\",\n      tags: [:controller, :action]\n    )","ref":"Telemetry.Metrics.html#counter/2-example"},{"type":"function","title":"Telemetry.Metrics.distribution/2","doc":"Returns a definition of distribution metric.\n\nDistribution metric builds a histogram of selected measurement's values. It is up to the reporter\nto decide how the boundaries of the distribution buckets are configured - via `:reporter_options`,\nconfiguration of the aggregating system, or other means.\n\nSee the [\"Metrics\"](#module-metrics) section in the top-level documentation of this module for more\ninformation.","ref":"Telemetry.Metrics.html#distribution/2"},{"type":"function","title":"Example - Telemetry.Metrics.distribution/2","doc":"distribution(\n      \"http.request.duration\",\n      tags: [:controller, :action],\n    )","ref":"Telemetry.Metrics.html#distribution/2-example"},{"type":"function","title":"Telemetry.Metrics.last_value/2","doc":"Returns a definition of last value metric.\n\nLast value keeps track of the selected measurement found in the most recent event.\n\nSee the [\"Metrics\"](#module-metrics) section in the top-level documentation of this module for more\ninformation.","ref":"Telemetry.Metrics.html#last_value/2"},{"type":"function","title":"Example - Telemetry.Metrics.last_value/2","doc":"last_value(\n      \"vm.memory.total\",\n      description: \"Total amount of memory allocated by the Erlang VM\", unit: :byte\n    )","ref":"Telemetry.Metrics.html#last_value/2-example"},{"type":"function","title":"Telemetry.Metrics.sum/2","doc":"Returns a definition of sum metric.\n\nSum metric keeps track of the sum of selected measurement's values carried by specific events.\n\nSee the [\"Metrics\"](#module-metrics) section in the top-level documentation of this module for more\ninformation.","ref":"Telemetry.Metrics.html#sum/2"},{"type":"function","title":"Example - Telemetry.Metrics.sum/2","doc":"sum(\n      \"user.session_count\",\n      event_name: \"user.session_count\",\n      measurement: :delta,\n      tags: [:role]\n    )","ref":"Telemetry.Metrics.html#sum/2-example"},{"type":"function","title":"Telemetry.Metrics.summary/2","doc":"Returns a definition of summary metric.\n\nThis metric aggregates measurement's values into statistics, e.g. minimum and maximum, mean, or\npercentiles. It is up to the reporter to decide which statistics exactly are exposed.\n\nSee the [\"Metrics\"](#module-metrics) section in the top-level documentation of this module for more\ninformation.","ref":"Telemetry.Metrics.html#summary/2"},{"type":"function","title":"Example - Telemetry.Metrics.summary/2","doc":"summary(\n      \"db.query.duration\",\n      tags: [:table],\n      unit: {:native, :millisecond}\n    )","ref":"Telemetry.Metrics.html#summary/2-example"},{"type":"type","title":"Telemetry.Metrics.byte_unit/0","doc":"","ref":"Telemetry.Metrics.html#t:byte_unit/0"},{"type":"type","title":"Telemetry.Metrics.byte_unit_conversion/0","doc":"","ref":"Telemetry.Metrics.html#t:byte_unit_conversion/0"},{"type":"type","title":"Telemetry.Metrics.description/0","doc":"","ref":"Telemetry.Metrics.html#t:description/0"},{"type":"type","title":"Telemetry.Metrics.drop/0","doc":"","ref":"Telemetry.Metrics.html#t:drop/0"},{"type":"type","title":"Telemetry.Metrics.keep/0","doc":"","ref":"Telemetry.Metrics.html#t:keep/0"},{"type":"type","title":"Telemetry.Metrics.measurement/0","doc":"","ref":"Telemetry.Metrics.html#t:measurement/0"},{"type":"type","title":"Telemetry.Metrics.metric_name/0","doc":"The name of the metric, either as string or a list of atoms.","ref":"Telemetry.Metrics.html#t:metric_name/0"},{"type":"type","title":"Telemetry.Metrics.metric_option/0","doc":"","ref":"Telemetry.Metrics.html#t:metric_option/0"},{"type":"type","title":"Telemetry.Metrics.metric_options/0","doc":"","ref":"Telemetry.Metrics.html#t:metric_options/0"},{"type":"type","title":"Telemetry.Metrics.normalized_metric_name/0","doc":"The name of the metric represented as a list of atoms.","ref":"Telemetry.Metrics.html#t:normalized_metric_name/0"},{"type":"type","title":"Telemetry.Metrics.reporter_options/0","doc":"","ref":"Telemetry.Metrics.html#t:reporter_options/0"},{"type":"type","title":"Telemetry.Metrics.t/0","doc":"One of the base metric definitions.","ref":"Telemetry.Metrics.html#t:t/0"},{"type":"type","title":"Telemetry.Metrics.tag/0","doc":"","ref":"Telemetry.Metrics.html#t:tag/0"},{"type":"type","title":"Telemetry.Metrics.tag_values/0","doc":"","ref":"Telemetry.Metrics.html#t:tag_values/0"},{"type":"type","title":"Telemetry.Metrics.tags/0","doc":"","ref":"Telemetry.Metrics.html#t:tags/0"},{"type":"type","title":"Telemetry.Metrics.time_unit/0","doc":"","ref":"Telemetry.Metrics.html#t:time_unit/0"},{"type":"type","title":"Telemetry.Metrics.time_unit_conversion/0","doc":"","ref":"Telemetry.Metrics.html#t:time_unit_conversion/0"},{"type":"type","title":"Telemetry.Metrics.unit/0","doc":"","ref":"Telemetry.Metrics.html#t:unit/0"},{"type":"module","title":"Telemetry.Metrics.ConsoleReporter","doc":"A reporter that prints events and metrics to the terminal.\n\nThis is useful for debugging and discovering all available\nmeasurements and metadata in an event.\n\nFor example, imagine the given metrics:\n\n    metrics = [\n      last_value(\"vm.memory.binary\", unit: :byte),\n      counter(\"vm.memory.total\")\n    ]\n\nA console reporter can be started as a child of your supervision tree as:\n\n    {Telemetry.Metrics.ConsoleReporter, metrics: metrics}\n\nNow when the \"vm.memory\" telemetry event is dispatched, we will see\nreports like this:\n\n    [Telemetry.Metrics.ConsoleReporter] Got new event!\n    Event name: vm.memory\n    All measurements: %{binary: 100, total: 200}\n    All metadata: %{}\n\n    Metric measurement: :binary (last_value)\n    With value: 100 byte\n    And tag values: %{}\n\n    Metric measurement: :total (counter)\n    With value: 200\n    And tag values: %{}\n\nIn other words, every time there is an event for any of the registered\nmetrics, it prints the event measurement and metadata, and then it prints\ninformation about each metric to the user.","ref":"Telemetry.Metrics.ConsoleReporter.html"},{"type":"function","title":"Telemetry.Metrics.ConsoleReporter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Telemetry.Metrics.ConsoleReporter.html#child_spec/1"},{"type":"function","title":"Telemetry.Metrics.ConsoleReporter.start_link/1","doc":"","ref":"Telemetry.Metrics.ConsoleReporter.html#start_link/1"},{"type":"module","title":"Telemetry.Metrics.Counter","doc":"Defines a specification of counter metric.","ref":"Telemetry.Metrics.Counter.html"},{"type":"type","title":"Telemetry.Metrics.Counter.t/0","doc":"","ref":"Telemetry.Metrics.Counter.html#t:t/0"},{"type":"module","title":"Telemetry.Metrics.Distribution","doc":"Defines a specification of distribution metric.","ref":"Telemetry.Metrics.Distribution.html"},{"type":"type","title":"Telemetry.Metrics.Distribution.t/0","doc":"","ref":"Telemetry.Metrics.Distribution.html#t:t/0"},{"type":"module","title":"Telemetry.Metrics.LastValue","doc":"Defines a specification of last value metric.","ref":"Telemetry.Metrics.LastValue.html"},{"type":"type","title":"Telemetry.Metrics.LastValue.t/0","doc":"","ref":"Telemetry.Metrics.LastValue.html#t:t/0"},{"type":"module","title":"Telemetry.Metrics.Sum","doc":"Defines a specification of sum metric.","ref":"Telemetry.Metrics.Sum.html"},{"type":"type","title":"Telemetry.Metrics.Sum.t/0","doc":"","ref":"Telemetry.Metrics.Sum.html#t:t/0"},{"type":"module","title":"Telemetry.Metrics.Summary","doc":"Defines a specification of summary metric.","ref":"Telemetry.Metrics.Summary.html"},{"type":"type","title":"Telemetry.Metrics.Summary.t/0","doc":"","ref":"Telemetry.Metrics.Summary.html#t:t/0"},{"type":"extras","title":"Rationale","doc":"# Rationale\n\nThe design proposed by Telemetry.Metrics might look controversial - unlike most of the libraries\navailable on the BEAM, it doesn't aggregate metrics itself, it merely defines what users should\nexpect when using the reporters.\n\nIf Telemetry.Metrics would aggregate metrics, the way those aggregations work would be imposed\non the system where the metrics are published to. For example, counters in StatsD are reset on\nevery flush and can be decremented, whereas counters in Prometheus are monotonically increasing.\nTelemetry.Metrics doesn't focus on those details - instead, it describes what the end user,\noperator, expects to see when using the metric of particular type. This implies that in most\ncases aggregated metrics won't be visible inside the BEAM, but in exchange aggregations can be\nimplemented in a way that makes most sense for particular system.\n\nFinally, one could also implement an in-VM \"reporter\" which would aggregate the metrics and expose\nthem inside the BEAM. When there is a need to swap the reporters, and if both reporters are\nfollowing the metric types specification, then the end result of aggregation is the same,\nregardless of the backend system in use.","ref":"rationale.html"},{"type":"extras","title":"Writing Reporters","doc":"# Writing Reporters\n\nReporters are a crucial part of Telemetry.Metrics ecosystem. Without them, metric\ndefinitions are merely... definitions. This guide aims to help in writing reporters\nin a proper way.\n\n> Before writing the reporter for your favourite monitoring system, make sure that one isn't\n> already available on Hex.pm - it might make sense to contribute and improve the existing\n> solution than starting from scratch.\n\nLet's get started!","ref":"writing_reporters.html"},{"type":"extras","title":"Specification - Writing Reporters","doc":"- Reporters MUST accept a list of metric definitions as input when being started\n  - Reporters MUST attach handlers to events contained in these definitions\n  - Reporters MUST extract the measurement and selected tags specified by the metric definitions\n  - Reporters SHOULD handle events in a way that makes sense for whatever it\n    is publishing to\n  - Reporters MUST clean up on exit by detaching all event handlers they have created\n  - Reporters MUST honor `keep` recording rule functions\n  - Reporters MUST skip events with missing or invalid measurements or tags\n\n#","ref":"writing_reporters.html#specification"},{"type":"extras","title":"Accepting Metric Definitions as Input - Writing Reporters","doc":"This one is quite easy - you need to give your users a way to actually tell you what metrics\nthey want to track. It's essential to give users an option to provide metric definitions\nat runtime (e.g. when their application starts). For example, let's say you're building a\n`PigeonReporter`.\n\nIf the reporter was process-based, you could provide a `start_link/1` function that accepts\na list of metric definitions:\n\n```elixir\nmetrics = [\n  counter(\"...\"),\n  last_value(\"...\"),\n  summary(\"...\")\n]\n\nPigeonReporter.start_link(metrics: metrics)\n```\n\nIf the reporter doesn't support metrics of particular type, it may either:\n\n  1. Log a warning and discard the metric\n  2. Log a warning and convert the metric to an equivalent type. For example, a reporter\n     may convert an histogram into a summary or simpler metric in case it is not supported\n\nWe recommend all reporters to include a summary table of which metrics are supported and\ntheir equivalents on the adapter terminology.\n\nReporter-specific options for individual metrics may be passed on the `:reporter_options`\nkey of the metric definitions. These options can be used to define options such as sample\nrates, percentiles, rates, etc. Reporters should validate any options they accept and\nprovide useful exception messages.\n\n#","ref":"writing_reporters.html#accepting-metric-definitions-as-input"},{"type":"extras","title":"Attaching event handlers - Writing Reporters","doc":"Event handlers are attached using `:telemetry.attach/4` function. To reduce overhead of\ninstalling many event handlers, you can install a single handler for multiple metrics\nbased on the same event but note that any exception will cause all metrics on under that\nhandler. You can achieve this by grouping the metrics by event name:\n\n```elixir\nEnum.group_by(metrics, & &1.event_name)\n```\n\nNote that handler IDs need to be unique - you can generate completely random blobs of\ndata, or use something that you know needs to be unique anyway, e.g. some combination\nof reporter name, event name, and something which is different for multiple instances\nof the same reporter (PID is a good choice as most reporters should be backed by a process):\n\n```elixir\nid = {PigeonReporter, metric.event_name, self()}\n```\n\nPutting it all together:\n\n```elixir\nfor {event, metrics} <- Enum.group_by(metrics, & &1.event_name) do\n  id = {__MODULE__, event, self()}\n  :telemetry.attach(id, event, &handle_event/4, metrics)\nend\n```\n\n#","ref":"writing_reporters.html#attaching-event-handlers"},{"type":"extras","title":"Reacting to events - Writing Reporters","doc":"When consuming events, there are five steps to take into account:\n\n1. If a `keep` recording rule function has been provided, the reporter MUST record the metric\n   only if the function returns `true`.\n\n2. Extract event measurements from the event. Measurements are optional, so we must skip\n   reporting that particular measurement if it is not available;\n\n3. Extract all the relevant tags from the event metadata (if they are supported by the reporter);\n\n4. Implement the logic specific to the reporter;\n\n5. How to react to errors. One option is to let the `handle_event/4` callback fail, but\nthat means we will no longer listen to any future event. Another option is to rescue any\nerror and log them. That's the approach we will take in this example. However, be careful!\nIf an event always contains bad data, then we will log an error every time it is emitted;\n\nLet's see a simple handler implementation that takes all of those four items into account:\n\n```elixir\ndef handle_event(_event_name, measurements, metadata, metrics) do\n  for metric <- metrics do\n    try do\n      if measurement = keep?(metric, metadata) && extract_measurement(metric, measurements) do\n        tags = extract_tags(metric, metadata)\n\n        # record and send\n      end\n    rescue\n      e ->\n        Logger.error(\"Could not format metric #{inspect metric}\")\n        Logger.error(Exception.format(:error, e, __STACKTRACE__))\n    end\n  end\nend\n```\n\nThe implementation of `keep?/2` might look like:\n\n```elixir\ndefp keep?(%{keep: keep}, metadata) when keep != nil, do: keep.(metadata)\ndefp keep?(_metric, _metadata), do: true\n```\n\nThe implementation of `extract_measurement/2` might look as follows:\n\n```elixir\ndef extract_measurement(metric, measurements) do\n  case metric.measurement do\n    fun when is_function(fun, 1) -> fun.(measurements)\n    key -> measurements[key]\n  end\nend\n```\n\nSince `:measurement` in the metric definition can be both an arbitrary term (to be used\nas key to fetch the measurement) or a function, we need to handle both cases.\n\n> Note: Telemetry.Metrics can't guarantee that the extracted measurement's value is a number.\n> Each reporter can handle this scenario properly, either by logging a warning, detaching\n> the handler etc.\n\nWe also need to implement the `extract_tags/2` function:\n\n```elixir\ndef extract_tags(metric, metadata) do\n  tag_values = metric.tag_values.(metadata)\n  Map.take(tag_values, metric.tags)\nend\n```\n\nFirst we need to apply last-minute transformation to the metadata using the `:tag_values`\nfunction, then we fetch all transformed metadata, ignoring any tag that may not be available.\n\n#","ref":"writing_reporters.html#reacting-to-events"},{"type":"extras","title":"Detaching Handlers on Termination - Writing Reporters","doc":"To leave the system in a clean state, the reporter must detach the event handlers it installed\nwhen it's being stopped or terminated unexpectedely. This can be done by trapping exits in the\n`init` function and implementing the terminate callback, or having a dedicated process\nresponsible only for the cleanup (e.g. by using monitors).","ref":"writing_reporters.html#detaching-handlers-on-termination"},{"type":"extras","title":"Documentation - Writing Reporters","doc":"It's extremely important that reporters document how `Telemetry.Metrics` metric types, names,\nand tags are translated to metric types and identifiers in the system they publish metrics to\n(this is particularly important for a summary metric which is broadly defined). They should also\ndocument if some metric types are not supported at all.","ref":"writing_reporters.html#documentation"},{"type":"extras","title":"Examples - Writing Reporters","doc":"This repository ships with a `Telemetry.Metrics.ConsoleReporter` that prints data to the\nterminal as an example. Official reporters can be found in the [BEAM Telemetry Github Organization](https://github.com/beam-telemetry). You may search for other reporters on hex.pm.","ref":"writing_reporters.html#examples"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}